# 支持动态切换界面语言

之前下载器的界面文本只在初始化时设置一次，之后无法根据语言动态变化。

例如之前的代码：

```js
const html = `<span>${lang.transl('_提示')}</span>`
```

如果此时的语言是简体中文，结果如下：

```html
<span>提示</span>
```

这种方法存在问题：如果用户切换了语言（例如切换为 English），本程序无法让文本动态变化。

现在我在需要根据语言动态切换文本的元素上添加了标记，这样语言模块就可以找到这些元素，并在用户切换语言时重设其文本。

# 步骤

1. 在需要动态切换语言的元素上添加特定标记（属性）
2. 使用 `lang.register()` 方法注册这个元素

这样，当语言发生变化时，本程序就会自动更新元素的文本。

# 标记列表

- `data-xztext`
- `data-xztextargs`
- `data-xztip`
- `data-xztitle`
- `data-xzplaceholder`

## 设置文本内容

在文本的容器元素上添加 `data-xztext` 属性：

```js
const html = `<span data-xztext="_提示"></span>`
const el = body.append(html)
lang.register(el)
```

`Lang` 模块会设置容器的 innerHTML：

```html
<span data-xztext="_提示">提示</span>
```

## 设置带参数的文本内容

有些文本有占位符，可以传入参数，例如：

```js
lang.transl('_转换任务提示', '6')
// 结果
'正在转换 6 个文件'
```

对于这种情况，需要使用 `data-xztextargs` 属性，属性值是所有参数使用 `,` 合并后的字符串：

```js
const html = `<span data-xztextargs="_转换任务提示,6"></span>`
```

## 设置提示内容

在需要设置提示内容的元素上添加 `data-xztip` 属性：

```js
const html = `<span data-xztip="_提示"></span>`
```

语言模块会设置容器的 `data-tip` 属性：


```html
<span data-xztip="_提示" data-tip="提示"></span>
```

## 设置 title 属性

在需要设置提示内容的元素上添加 `data-xztitle` 属性：

```js
const html = `<span data-xztitle="_标题"></span>`
```

语言模块会设置容器的 `title` 属性：


```html
<span data-xztitle="_标题" title="标题"></span>
```

## 设置 placeholder 属性

类似设置 title 属性，只是 `title` 换成了 `placeholder`。

```js
const html = `<span data-xzplaceholder="_提示"></span>`
```

# 文本发生变化时的处理

**调用 `lang.updateText()` 方法。**

有些元素的文本可能会动态变化，例如一个表示下载状态的元素：

```html
<span data-xztext="_下载中">下载中</span>
```

当下载完成时变化为：

```html
<span data-xztext="_下载完毕">下载完毕</span>
```

如果文本需要由 `Lang` 模块进行管理，则只需要调用 `lang.updateText()` 方法，示例：

```js
lang.updateText(el, '_下载完毕')

// 如果文本带参数，则添加到后面即可
lang.updateText(el, '_转换任务提示', '6')
```

调用此方法之后，`Lang` 模块会自动更新元素的文本和标记。例如：

```js
el.dataset.xztext = '_下载完毕'
el.innerHTML = '下载完毕'
```

## 清空文本时的处理

如果一个元素的文本被清空，只需要在 `lang.updateText()` 方法里传递空的文本标记：

```js
lang.updateText(el, '')
```

`Lang` 模块会自动将文本设置为空字符串 `''`，并且自动删除标记。例如：

```js
delete el.dataset.xztext
el.innerHTML = ''
```