# 尝试转换动图时避免页面卡顿，失败了

2022/08/31

测试的动图作品：

https://www.pixiv.net/artworks/100875971

21.3 MB，123 张图片，每张分辨率 640*360，177 KB。

未完成的 Worker 代码保存在：`src\ts\ConvertUgoira\convertUgoiraImages.worker.ts`

## GIF 图片

转换 GIF 图片时，前台页面渲染不会卡住，所以不需要优化。

## WebM 视频

转换时会导致前台页面停止渲染。

从压缩包下载完毕到转换完成，每个流程花费的时间：

1. 提取每一张图片: 815.622802734375 ms
2. 生成每一张图片的 DataURL: 5825.115966796875 ms
3. 编码 WebM: 315.064697265625 ms

可以看到，为每一张图片生成 DataURL 花费了大量的时间，这导致页面渲染卡住了。

为什么这部分这么慢？我猜测可能是因为 WebP 图片的编解码时间比较久。

我尝试把生成 DataURL 的任务转移到 Worker 里，但是失败了，因为图片需要被转码为 webp 格式，然后生成 DataURL，关键代码：

```js
// Worker
const canvas = new OffscreenCanvas(ev.data.width, ev.data.height)
const ctx = canvas.getContext('2d')!
ctx.drawImage(ev.data.ImageBitmap, 0, 0, ev.data.width, ev.data.height)

// Error!
canvas.toDataURL('image/webp', 0.9)
```

因为在 Worker 里，不能使用 `canvas.toDataURL` 方法（`OffscreenCanvas` 上没有 `toDataURL` 方法），所以此任务只能在前台执行。

其实有一个变通的方法可以得到 DataURL，出自：https://stackoverflow.com/questions/56494320/js-offscreencanvas-todataurl

```js
const blob = await canvas.convertToBlob({
  type: 'image/webp',
  quality: 0.9
})
const dataURL = new FileReaderSync().readAsDataURL(blob)
```

但是根据第二个回答的描述，在这个变通的方法中，`convertToBlob` 的执行效率很低，还不如在前台执行，转换的快一些。

----------------

WebM 想要避免卡顿还有另一个办法，就是直接向编码器传递 canvas 对象，让编码器去执行 `canvas.toDataURL('image/webp', 0.9)`（而不是由下载器执行）。

为什么这样就不会导致渲染卡住了呢？因为编码器在对每一帧进行编码时，使用了 setTimeout 定时器，让每一帧之间有一定的间隔，这样页面渲染就不会卡住。而下载器是直接循环进行编码的，中间没有间隔。

但是让编码器去编码，由于有定时器，导致编码时间变长了，增加了转换动图的时间。以这个动图为例，增加了 1  秒钟。所以我没有这么做。

## APNG 图片

转换时会导致前台页面停止渲染。

从压缩包下载完毕到转换完成，每个流程花费的时间：

1. 提取每一张图片: 749.477783203125 ms
2. 生成每一张图片的 PNG 格式 buffer: 1202.7880859375 ms
3. 编码 APNG 图片: 13561.01708984375 ms

这和 WebM 的情况不同，WebM 是下载器生成 DataURL 花费了大部分时间，而 APNG 是编码（第三方库）花费了大量时间。

我只能优化下载器负责的部分，主要是第二部分。这部分成功的转移到了 Worker 里，但是由于其总时间也就 1 秒（其实优化后更短了），感知不明显，所以我并没有实装它。

```js
// Worker
ctx.drawImage(ev.data.ImageBitmap, 0, 0, ev.data.width, ev.data.height)
// 从画布获取图像绘制后的 Uint8ClampedArray buffer
const buffer = ctx.getImageData(0, 0, ev.data.width, ev.data.height).data.buffer
```

---------------------------

## 优化前两个步骤

虽然避免页面渲染卡顿的优化没有成功，不过我依然优化了前两个步骤。

之前，下载器会把图片从 blob 转换为 img，现在改为从 blob 转换为 ImageBitmap。

在第一步，可以减少加载图片的等待时间；

在第二步，在 canvas 上绘制 ImageBitmap 并获取图像数据，比之前绘制 img 的方式更加高效。

（对第三步没有影响）

这个优化效果很玄学，在我刚修改完之后，节省的时间比较明显，如下：

### WebM 视频

| 步骤                       | 之前    | 现在    |
| -------------------------- | ------- | ------- |
| 1. 提取每一张图片          | 815 ms  | 225 ms  |
| 2. 生成每一张图片的 buffer | 5825 ms | 5427 ms |

前两步一共节省了大约 1 秒。


### APNG 图片

| 步骤                       | 之前    | 现在   |
| -------------------------- | ------- | ------ |
| 1. 提取每一张图片          | 749 ms  | 231 ms |
| 2. 生成每一张图片的 buffer | 1202 ms | 215 ms |

前两步一共节省了 1.5 秒。

### GIF 图片

GIF 图片 之前只有 2 个步骤：

1. 提取每一张图片: 769.348876953125 ms
2. 编码 GIF: 7415.05517578125 ms

现在优化之后有 3 个步骤：

1. 提取每一张图片: 900.493896484375 ms
2. 获取每一张图片的 ImageData: 136.63330078125 ms
3. 编码 GIF: 7117.5859375 ms

时间没怎么减少。

----------------

然而，当我修改完一会儿之后，我再次测试，发现第一步的时间反而比优化之前增加了。

第二步的优化效果依然还在。

很奇怪。

